/**
 * @cond skip
 * vim:syntax=doxygen
 * @endcond
 *
 *
@page module_doc

@section module_intro Introduction

FreeRADIUS uses a pluggable module infrastructure to extend the
basic functionality of the server.

Modules in 3.0 are dynamically linked at runtime using dlopen.

@defgroup module_safe module_safe - Internal API functions that are safe for use by modules
*/
